@@ -1,5 +1,6 @@
 """
-Parts of this implementation are taken from akaihola's snippet on stackoverflow:
+Parts of this implementation are taken from akaihola's snippet on
+stackoverflow:
 
   http://stackoverflow.com/questions/431628/how-to-combine-2-or-more-querysets-in-a-django-view
 
@@ -46,10 +47,10 @@ class SortedQuerySetChain(object):
         return sum(qs.count() for qs in self.querysets)
 
     def _sort_key(self, obj):
-        key = self.sort_keys.get(type(obj), None)
+        key = self.sort_keys.get(type(obj[0]), None)
         if key:
-            return key(obj)
-        return obj
+            return key(obj[0])
+        return obj[0]
 
     def __getitem__(self, idx):
         if isinstance(idx, slice):
@@ -61,33 +62,49 @@ class SortedQuerySetChain(object):
             return islice(self._all(), idx, idx + 1).next()
 
     def _all(self):
-        next = []
-        querysets = list(self.querysets)
-        i = 0
-        j = i + self.chunk_size
-        to_remove = []
+        chunk_size = self.chunk_size
+
+        def contains(list, qs_idx):
+            """Checks if the list contains a qs_idx."""
+            for x in list:
+                if x[1] == qs_idx:
+                    return True
+            return False
+
+        class QSPointer(object):
+
+            def __init__(self, qs):
+                self.qs = qs
+                self.next = 0  # index of next unread item
+
+        querysets = dict((idx, QSPointer(qs)) for
+                         idx, qs in enumerate(self.querysets))
+        next_results = []  # a list of tuples: (element from qs, qs idx)
+        qs_to_remove = []
         while True:
-            for q in querysets[:]:
-                chunk = list(islice(q, i, j))
-                if not chunk:
-                    to_remove.append(q)
-                else:
-                    next.extend(chunk)
-            if to_remove:
-                for q in to_remove:
-                    querysets.remove(q)
-                del to_remove[:]
+            next_modified = False
+            for idx in querysets:
+                qs_pointer = querysets[idx]
+                if not contains(next_results, idx):
+                    end = qs_pointer.next + chunk_size
+                    chunk = qs_pointer.qs[qs_pointer.next:end]
+                    if not chunk:
+                        qs_to_remove.append(idx)
+                    else:
+                        querysets[idx].next = end
+                        next_results.extend([(item, idx) for item in chunk])
+                        next_modified = True
 
-            if next:
-                next.sort(key=self._sort_key, reverse=self.reverse)
-                for n in next:
-                    yield n
-                del next[:]
+            while qs_to_remove:
+                del querysets[qs_to_remove.pop()]
 
+            if next_modified:
+                next_results.sort(key=self._sort_key, reverse=not self.reverse)
+
+            if next_results:
+                yield next_results.pop()[0]
             else:
                 break
 
-            i, j = j, j + self.chunk_size
-
 
 __all__ = ['SortedQuerySetChain']
